﻿using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;

namespace HeuristicLab.BioBoost.Utils {

  /// <summary>
  /// Wraps a field into a local variable inside the constructor. This evades
  /// object graph traversal as the actual value is no longer discoverable by
  /// reflection. Use with Caution!
  /// </summary>
  /// <typeparam name="T">The Type of the value to be wrapped.</typeparam>
  [StorableClass]
  public class Closure<T> : IDeepCloneable {

    public enum Encapsulation { Cloned, Referenced };

    [Storable]
    private readonly Encapsulation encapsulation;

    [Storable]
    public T Value { get { return Getter(); } set { Setter(value); } }

    public Func<T> Getter;
    public Action<T> Setter;

    [StorableConstructor]
    protected Closure(bool isDeserializing) {
      T value = default(T);
      Getter = () => value;
      Setter = v => value = v;
    }
    protected Closure(Closure<T> orig, Cloner cloner) {
      T value = default(T);
      Getter = () => value;
      Setter = v => value = v;
      encapsulation = orig.encapsulation;
      cloner.RegisterClonedObject(orig, this);
      if (encapsulation == Encapsulation.Referenced) {
        Value = orig.Value;
      } else {
        var cloneable = orig.Value as IDeepCloneable;
        if (cloneable != null)
          Value = (T) cloner.Clone(cloneable);
        else
          Value = orig.Value;
      }
    }
    public Closure(Encapsulation encapsulation) {
      this.encapsulation = encapsulation;
      T value = default(T);
      Getter = () => value;
      Setter = v => value = v;
    }

    public object Clone() {
      return Clone(new Cloner());
    }

    public IDeepCloneable Clone(Cloner cloner) {
      return new Closure<T>(this, cloner);
    }
  }
}
