﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Mono;

namespace HeuristicLab.BioBoost.Utils {

  public class AnonymousEqualityComparer<T> : EqualityComparer<T> {

    public Func<T, T, bool> ComparerFunction { get; private set; } 

    public AnonymousEqualityComparer(Func<T, T, bool> comparer) { ComparerFunction = comparer; }

    public override bool Equals(T x, T y) { return ComparerFunction(x, y); }

    public override int GetHashCode(T obj) { return obj == null ? 0 : obj.GetHashCode(); }
  }

  public static class ListExtensions {

    public static void AddRange<T>(this IList<T> l, IEnumerable<T> elements) { foreach (var e in elements) l.Add(e); }

    public static IEnumerable<T> RemoveAtIndices<T>(this IEnumerable<T> l, IEnumerable<int> orderedIndices) {
      int i = 0;
      var listIt = l.GetEnumerator();
      var idxIt = orderedIndices.GetEnumerator();
      bool done = !idxIt.MoveNext();
      while (listIt.MoveNext()) {
        Debug.Assert(done || idxIt.Current >= i);
        if (!done && idxIt.Current == i) {
          done = !idxIt.MoveNext();
        } else {
          yield return listIt.Current;
        }
        i++;
      }
    }

    public static Queue<T> ToQueue<T>(this IEnumerable<T> l) { return new Queue<T>(l); }

    public static IEnumerable<TResult> Select<T, TResult>(this IEnumerable<T> l, TResult defaultValue, Func<T, TResult> f) {
      return l.Select(x => x == null ? defaultValue : f(x));
    }

    public static DoubleArray ToDoubleArray(this IEnumerable<double> t) { return new DoubleArray(t.ToArray()); }
    public static StringArray ToStringArray(this IEnumerable<string> t) { return new StringArray(t.ToArray()); }

    public static IEnumerable<IGrouping<TKey, TResult>> GroupBy<T, TKey, TResult>(this IEnumerable<T> source,
      Func<T, TKey> keySelector, Func<T, TResult> resultSelector, Func<TKey, TKey, bool> equalityComparer) {
      return source.GroupBy(keySelector, resultSelector, new AnonymousEqualityComparer<TKey>(equalityComparer));
    }

    public static HashSet<T> ToHashSet<T>(this IEnumerable<T> values) { return new HashSet<T>(values); }

    public static TResult FuncOverIndices<T, TResult>(this IEnumerable<T> values, ISet<int> indices, Func<IEnumerable<T>, TResult> func) {
      return func(values.Select((v, i) => new {v, i}).Where(p => indices.Contains(p.i)).Select(p => p.v));
    }

    public static IDictionary<int, int[]> SourceIndices(this IEnumerable<int> targets) {
      return targets
        .Select((t, idx) => new {t, idx})
        .GroupBy(ti => ti.t)
        .ToDictionary(
          tis => tis.Key,
          tis => tis.Select(ti => ti.idx).ToArray());
    } 
  }

}
